//freeze;

// generic type for diagonal quadratic forms
declare type RngWittQF[RngWittQFElt];// : Rng;
declare attributes RngWittQF: base_ring, is_explicit, explicit_model;
declare attributes RngWittQFElt: form, parent;
declare verbose WR_verbosity, 3;

import "rationals/quadratic_forms.m" : is_torsion_QQ;
import "number_fields/quadratic_forms.m" : is_torsion_NF;
import "rational_global_functions/quadratic_forms.m" : is_torsion_RFF;
import "global_function_fields/quadratic_forms.m" : is_torsion_GFF;
import "real_numbers/quadratic_forms.m" : is_torsion_RR;
import "complex_numbers/quadratic_forms.m" : is_torsion_CC;
import "finite_fields/finite_field.m" : is_torsion_FF;

// printing
intrinsic Print( W::RngWittQF, L::MonStgElt )
{
Prints W
}
    if L eq "Maximal" and W`is_explicit then
        printf "Witt ring of %o, isomorphic to %o", W`base_ring, W`explicit_model;
    else
        printf "Witt ring of %o", W`base_ring;
    end if;
end intrinsic;

intrinsic Print( _q::RngWittQFElt )
{
Prints _q
}
    printf "[%o]", _q`form;
end intrinsic;

// constructors
intrinsic WittRing( F::Fld ) -> RngWittQF
{
Constructs the Witt ring of similarity classes of quadratic forms over F.
}
    W := New(RngWittQF);
    W`base_ring := F;
    W`is_explicit := false;
    return W;
end intrinsic;

intrinsic IsCoercible( W::RngWittQF, q::.) -> BoolElt, .
{
Return whether q is coercible into W and the result if so
}
    if ISA(Type(q), DiagQuadFrm) then
        // conversion of a diagonal quadratic form
        if BaseRing(q) ne BaseRing(W) then
            return false, "1 Illegal coercion!";
        end if;
        require IsNondegenerate(q): "The quadratic form must not be degenerate!";
        _q := New(RngWittQFElt);
        _q`form := DiagonalQuadraticForm( BaseRing(W), q`coeffs );
        _q`parent := W;
        return true, _q;
/*    elif Type(q) eq ModTupFld then
        if not IsPolarSpace(q) then
            return false, "Illegal coercion!";
        end if;
        if PolarSpaceType(q) ne "quadratic space" then
            return false, "Illegal coercion!";
        end if;
        M := QuadraticFormMatrix(q);
        D := Diagonalization( 1/2*(M+Transpose(M)) );
        n := Ncols(D);
        K := BaseRing(W);
        q1 := DiagonalQuadraticForm(K, [ D[j,j] : j in [1..n] ]);
        _q := New(RngWittQFElt);
        _q`form := DiagonalQuadraticForm( BaseRing(W), q1`coeffs );
        _q`parent := W;
        return true, _q; */
    elif ISA(Type(q), RngWittQFElt) then 
        // trivial conversion: element of Witt ring to itself
        if BaseRing(q) ne BaseRing(W) then
            return false, "2 Illegal coercion!";
        end if;
        return true, q;
    elif ISA(Type(q), RngElt) then
        // conversion of a scalar
        t, a := IsCoercible(BaseRing(W), q);
        if not t then
            return false, "3 Illegal coercion!";
        end if;
        _q := New(RngWittQFElt);
        _q`form := DiagonalQuadraticForm( BaseRing(W), [q] );
        _q`parent := W;
        return true, _q;
    else
        return false, "4 Illegal coercion!";
    end if;
end intrinsic;

// invariants
intrinsic Parent( _q::RngWittQFElt ) -> RngWittQF
{
Parent of _q
}
    return _q`parent;
end intrinsic;

intrinsic BaseRing( W::RngWittQF ) -> Rng
{
Returns the base ring of W
}
    return W`base_ring;
end intrinsic;

intrinsic BaseField(  W::RngWittQF ) -> Rng
{ " } //"
    return W`base_ring;
end intrinsic;

intrinsic BaseRing( _q::RngWittQFElt ) -> Rng
{
Returns the base ring of _q
}
    return BaseRing(Parent(_q));
end intrinsic;

intrinsic BaseField(  _q::RngWittQFElt ) -> Rng
{ " } //"
    return BaseRing(Parent(_q));
end intrinsic;

intrinsic '#'( W::RngWittQF ) -> .
{
The cardinality of the Witt ring W
}
    K := BaseRing(W);
    T := Category(K);
    t := ISA(T, FldFin);
    t or:= ISA(T, FldRat);
    t or:= ISA(T, FldNum);
    t or:= ISA(T, FldFunRat) and ISA(Category(ConstantField(K)), FldFin);
    t or:= ISA(T, FldFun) and ISA(Category(ConstantField(K)), FldFin);
    t or:= ISA(T, FldRe);
    t or:= ISA(T, FldCom);
    require t: "The base field is not supported!";
    return W`is_explicit select #(W`explicit_model) else Infinity();
end intrinsic;


intrinsic 'eq'( _q1::RngWittQFElt, _q2::RngWittQFElt ) -> BoolElt
{
Checks equality of Witt ring elements
}
    require BaseRing(_q1) eq BaseRing(_q2): "Elements must be belong to the same Witt ring!";
    q := _q1`form - _q2`form;
    return IsHyperbolic(q);
end intrinsic;

intrinsic '+'( _q1::RngWittQFElt, _q2::RngWittQFElt ) -> RngWittQFElt
{
Sum of Witt ring elements
}
    require BaseRing(_q1) eq BaseRing(_q2): "Elements must be belong to the same Witt ring!";
    q := _q1`form + _q2`form;
    return Parent(_q1) ! q;
end intrinsic;

intrinsic '-'( _q1::RngWittQFElt, _q2::RngWittQFElt ) -> RngWittQFElt
{
Difference of Witt ring elements
}
    require BaseRing(_q1) eq BaseRing(_q2): "Elements must be belong to the same Witt ring!";
    q := _q1`form - _q2`form;
    return Parent(_q1) ! q;
end intrinsic;

intrinsic '-'( _q::RngWittQFElt ) -> RngWittQFElt
{
Negation of a Witt ring element
}
    q1 := DiagonalQuadraticForm( BaseRing(_q`form), [ -c : c in _q`form`coeffs ] );
    return Parent(_q) ! q1;
end intrinsic;

intrinsic '*'( _q1::RngWittQFElt, _q2::RngWittQFElt ) -> RngWittQFElt
{
Product of Witt ring elements
}
    require BaseRing(_q1) eq BaseRing(_q2): "Elements must be belong to the same Witt ring!";
    q := TensorProduct( _q1`form, _q2`form );
    return Parent(_q1) ! q;
end intrinsic;

// integer multiplication (not very efficient implementation)
intrinsic '*'( n::RngIntElt, q:RngWittQFElt ) -> RngWittQFElt
{
Product of a Witt ring element by an integer, i.e. the n-fold sum q +...+ q
}
    W := Parent(q);
    // special case
    if IsZero(n) then
        return Zero(W);
    end if;
    // reduce to a positive multiplication
    if n lt 0 then
        q := -q;
        n := -n;
    end if;
    // perform standard binary multiplication
    r := Zero(W);   // the result
    b := q;         // auxiliary variable to be doubled
    while not IsZero(n) do
        // check the last binary digit and update r if needed
        if IsOdd(n) then
            r +:= b;
        end if;
        // double b
        b +:= b;
        // half n
        n := ShiftRight(n, 1);
    end while;
    return r;
end intrinsic;

intrinsic Zero( W::RngWittQF ) -> RngWittQFElt
{
Zero element of the Witt ring W, i.e. the class of hyperbolic forms.
}
    _q := New(RngWittQFElt);
    _q`form := DiagonalQuadraticForm( W`base_ring, [] );
    _q`parent := W;
    return _q;
end intrinsic;

intrinsic IsZero( _q::RngWittQFElt ) -> BoolElt
{
Checks if _q is the zero element in the Witt ring, i.e. if it is the calss of hyperbolic forms.
}
    return IsZero(Dimension(_q`form)) select true else IsHyperbolic( _q`form );
end intrinsic;

intrinsic One( W::RngWittQF ) -> RngWittQFElt
{
Unit element of the Witt ring W, i.e. the class of a 1-dimensional form (1).
}
    _q := New(RngWittQFElt);
    K := W`base_ring;
    _q`form := DiagonalQuadraticForm( K, [ One(K) ] );
    _q`parent := W;
    return _q;
end intrinsic;

intrinsic IsOne( _q::RngWittQFElt ) -> BoolElt
{
Checks if _q is the unit element in the Witt ring, i.e. if it is the calss of the 1-dimensional form (1).
}
    return _q eq One(Parent(_q));
end intrinsic;


//////////////////////////////////////////////////
// IsTorsion
intrinsic IsTorsion( _q::RngWittQFElt ) -> BoolElt
{
Check if the element _q of a Witt ring is torsion.
}
    K := BaseRing(_q);
    require IsField(K): "The base ring of K must be a field.";

    if ISA(Category(K), FldRe) then
        return is_torsion_RR(_q);
    elif ISA(Category(K), FldCom) then
        return is_torsion_CC(_q);
    elif ISA(Category(K), FldRat) then
        return is_torsion_QQ(_q);
    elif ISA(Category(K), FldFin) then
        return is_torsion_FF(_q);
    elif ISA(Category(K), FldNum) then
        return is_torsion_NF(_q);
    elif ISA(Category(K), FldFun) then
        if ISA(Category(ConstantField(K)), FldFin) then
            return is_torsion_GFF(_q);
        end if;
    elif ISA(Category(K), FldFunRat) then
        if ISA(Category(ConstantField(K)), FldFin) then
            return is_torsion_RFF(_q);
        end if;
    else
        require false: "The base field is not supported!";
    end if;
    
    return false;
end intrinsic;


//////////////////////////////////////////////////
// Height

intrinsic Height( K::Fld ) -> RngIntElt
{
This function computes the height h(K) of a field K,
i.e. exp( Wt(K) ), where Wt(K) is denotes the torsion subgroup
of the Witt ring of K.
}
    WK := WittRing(K);
    require WK`is_explicit: "Field K is not supported!";
    return 2*Level(K);
end intrinsic;


//////////////////////////////////////////////////
// AreWittEquivalent

intrinsic AreWittEquivalent( K::Fld, L::Fld ) -> BoolElt
{
Given two fields K and L, this function return true if their Witt rings are isomorphic and false otherwise.
}
    WK := WittRing(K);
    WL := WittRing(L);
    
    if WK`is_explicit ne WL`is_explicit then
        return false;
    elif WK`is_explicit then
        if Type(WK`explicit_model) ne Type(WL`explicit_model) then
            return false;
        end if;
        return WK`explicit_model eq WL`explicit_model;            // Risky! Will it always work? Is there a way to check ISOMORPHISM?
    else
        require false: "Not implemented for these types of fields!";
    end if;
end intrinsic;
