freeze;

// generic type for diagonal quadratic forms
declare type DiagQuadFrm;
declare attributes DiagQuadFrm: coeffs, dimension, base_ring, disc, primes, isovec;
declare verbose DQF_verbosity, 3;

import "../helpers/varia.m" : ZZ, QQ;

// basic contructor
intrinsic DiagonalQuadraticForm( K::Rng, D::[RngElt] ) -> DiagQuadFrm
{
Given a sequence of coefficients D = [d1, d2,..., dn], constructs a diagonal form\
q = ( d1, d2, ..., dn ) ~ d1*x1^2 + ... dn*xn^2.
}
    require forall{ d : d in D | IsCoercible(K, d) }: "Coefficients must be coercible to the base field.";
    q := New(DiagQuadFrm);
    q`coeffs := [ K!d : d in D ];
    q`dimension := #D;
    q`base_ring := K;
    return q;
end intrinsic;

intrinsic DiagonalQuadraticForm( D::[RngElt] ) -> DiagQuadFrm
{ " } //"
    return DiagonalQuadraticForm( Universe(D), D );
end intrinsic;


intrinsic HyperbolicForm( K::Rng, n::RngIntElt ) -> DiagQuadFrm
{
Construct a hyperbolic form over K of dimension n\\ 
(i.e. an orthogonal sum of n/2 hyperbolic planes).
}
    require IsEven(n): "The second argument must be even.";
    return DiagonalQuadraticForm(K, [K!1 : j in [1..n/2]] cat [K!-1: j in [1..n/2]]);
end intrinsic;

intrinsic OrthogonalSum( q1::DiagQuadFrm, q2::DiagQuadFrm ) -> DiagQuadFrm
{
Orthogonal sum of two diagonal quadratic forms.
}
    require q1`base_ring eq q2`base_ring: "Forms must be defined over the same ring.";
    
    return DiagonalQuadraticForm( q1`base_ring, q1`coeffs cat q2`coeffs );
end intrinsic;

intrinsic '+'( q1::DiagQuadFrm, q2::DiagQuadFrm ) -> DiagQuadFrm
{ " } //"
    return OrthogonalSum(q1, q2);
end intrinsic;

intrinsic '-'( q1::DiagQuadFrm, q2::DiagQuadFrm ) -> DiagQuadFrm
{
Computes an ``orthogonal difference'', i.e. a sum q1 + (-q2)
}
    return OrthogonalSum(q1, DiagonalQuadraticForm( q2`base_ring, [ -c : c in q2`coeffs ]));
end intrinsic;

intrinsic TensorProduct( q1::DiagQuadFrm, q2::DiagQuadFrm ) -> DiagQuadFrm
{
Tensor product of two diagonal quadratic forms.
}
    require q1`base_ring eq q2`base_ring: "Forms must be defined over the same ring.";
    
    return DiagonalQuadraticForm( q1`base_ring, [ c*d : c in q1`coeffs, d in q2`coeffs ] );
end intrinsic;

// constructor of a Pfister form
intrinsic PfisterForm( K::Rng, D::[RngElt] ) -> DiagQuadFrm
{
Given a sequence of coefficients D = [d1, d2,..., dn], constructs 
a Pfister form q = ( 1, d1, d2, ..., dn, d1*d2, d1*d3, ..., d1*d2*...*dn ).
}
    require forall{ d : d in D | IsCoercible(K, d) }: "Coefficients must be coercible to the base field.";
    q := DiagonalQuadraticForm( K, [One(K), D[1]] );
    for j in [2..#D] do
        q1 := DiagonalQuadraticForm( K, [One(K), D[j]] );
        q := TensorProduct( q, q1 );
    end for;
    return q;
end intrinsic;

intrinsic PfisterForm( D::[RngElt] ) -> DiagQuadFrm
{ " } //"
    PfisterForm( Universe(D), D );
end intrinsic;

intrinsic RandomShuffle( q::DiagQuadFrm : n := 3 ) -> DiagQuadFrm
{
Given a diagonal quadratic form q, this function creates a form q1, which is 2-conneced to q in n steps. By default n equals 3, but can be set to any positive integer.
}
    d := Dimension(q);
    C := q`coeffs;
    require d ge 2: "The dimension of the form must be at least 2.";
    for j in [1..n] do
        // pick two random coefficients
        repeat
            i1 := Random(1, d);
            i2 := Random(1, d);
        until i2 ne i1;
        a := C[i1];
        b := C[i2];
        a_ := a + b;
        if IsZero(a_) then
            continue;
        end if;
        b_ := a*b*a_;
        C[i1] := a_;
        C[i2] := b_;
    end for;

    return DiagonalQuadraticForm( BaseRing(q), C );
end intrinsic;

intrinsic Subform( q::DiagQuadFrm, I::[RngIntElt] ) -> DiagQuadFrm
{
Given a quadratic form q = (a_1,...,a_n) and a list of indices I, this function contsructs a a subform of q, having coefficients in I.
}
    n := Dimension(q);
    require forall{ i : i in I | i gt 0 and i le n } : "Incorrect list of indices.";
    K := BaseRing(q);
    C := Coefficients(q);
    q_ := DiagonalQuadraticForm(K, C[I]);
    if assigned q`primes then
        q_`primes := q`primes;
    end if;
    return q_;
end intrinsic;
