freeze;

import "../real_numbers/quadratic_forms.m" : is_isotropic_RR, is_hyperbolic_RR, anisotropic_dimension_RR;
import "../complex_numbers/quadratic_forms.m" : is_isotropic_CC, is_hyperbolic_CC, anisotropic_dimension_CC;
import "../rationals/quadratic_forms.m" : is_isotropic_QQ, is_hyperbolic_QQ, anisotropic_dimension_QQ;
import "../finite_fields/quadratic_forms.m" : is_isotropic_FF, is_hyperbolic_FF, anisotropic_dimension_FF;
import "../number_fields/quadratic_forms.m" : is_isotropic_NF, is_hyperbolic_NF, anisotropic_dimension_NF;
import "../global_function_fields/quadratic_forms.m" : is_isotropic_GFF, is_hyperbolic_GFF, anisotropic_dimension_GFF;
import "../rational_global_functions/quadratic_forms.m" : is_isotropic_RFF, is_hyperbolic_RFF, anisotropic_dimension_RFF;

//////////////////////////////////////////////////
// basic invariants

intrinsic Dimension( q::DiagQuadFrm ) -> RngIntElt
{
Dimension of q
}
    return q`dimension;
end intrinsic;

intrinsic BaseRing( q::DiagQuadFrm ) -> Rng
{
Returns the base ring of q
}
    return q`base_ring;
end intrinsic;

intrinsic BaseField( q::DiagQuadFrm ) -> Rng
{ " } //"
    return q`base_ring;
end intrinsic;

intrinsic Discriminant( q::DiagQuadFrm ) -> RngElt
{
Discriminant of a quadratic form q.
}
    if not assigned q`disc then
        d := Dimension(q);
        c := Integers() ! ((-1)^(d*(d-1)/2));
        q`disc := (BaseRing(q) ! c)* &* Coefficients(q);
    end if;
    return q`disc;
end intrinsic;

intrinsic Determinant( q::DiagQuadFrm ) -> RngElt
{
Discriminant of a quadratic form q.
}
    return &* Coefficients(q);
end intrinsic;


intrinsic IsNondegenerate( q::DiagQuadFrm ) -> BoolElt
{
Return true iff the form q is non-degenerate i.e. its determinant is non-zero.
}
    return forall{ c : c in q`coeffs | not IsZero(c) };
end intrinsic;

intrinsic IsNondegenerate( V::ModTupFld ) -> BoolElt
{ " } //"
    require IsPolarSpace(V): "Argument must be a quadratic space!";
    require PolarSpaceType(V) eq "quadratic space": "Argument must be a quadratic space!";
    K := BaseField(V);
    require Characteristic(K) ne 2: "Characteristic 2 is not supported";

    return not IsZero(Determinant(V));
end intrinsic;

intrinsic IsDegenerate( q::DiagQuadFrm ) -> BoolElt
{
Return true iff the form q is degenerate i.e. its determinant is zero.
}
    return not IsNondegenerate(q);
end intrinsic;

intrinsic IsDegenerate( V::ModTupFld ) -> BoolElt
{ " } //"
    return not IsNondegenerate(V);
end intrinsic;



//////////////////////////////////////////////////
// IsIsotropic

intrinsic IsIsotropic( q::DiagQuadFrm ) -> BoolElt
{
Checks if a quadratic form q is isotropic over its base ring. 
}
    K := q`base_ring;
    require IsField(K): "The base ring of K must be a field.";
    //require ISA(Category(K), FldFin) or ISA(Category(K), FldNum): "The base field is not supported!";

    if ISA(Category(K), FldRe) then
        return is_isotropic_RR(q);
    elif ISA(Category(K), FldCom) then
        return is_isotropic_CC(q);
    elif ISA(Category(K), FldRat) then
        return is_isotropic_QQ(q);
    elif ISA(Category(K), FldFin) then
        return is_isotropic_FF(q);
    elif ISA(Category(K), FldNum) then
        return is_isotropic_NF(q);
    elif ISA(Category(K), FldFun) then
        if IsGlobal(K) then
            return is_isotropic_GFF(q);
        end if;
    elif ISA(Category(K), FldFunRat) then
        if IsGlobal(K) then
            return is_isotropic_RFF(q);
        end if;
    else
        require false: "The base field is not supported!";
    end if;
    
    return false;
end intrinsic;

intrinsic IsIsotropic( V::ModTupFld ) -> BoolElt
{ " } //"
    require IsPolarSpace(V): "Argument must be a quadratic space!";
    require PolarSpaceType(V) eq "quadratic space": "Argument must be a quadratic space!";
    K := BaseField(V);
    require Characteristic(K) ne 2: "Characteristic 2 is not supported";
    
    q := Diagonalization(V);
    return IsIsotropic(q);
end intrinsic;

//////////////////////////////////////////////////
// IsAnisotropic

intrinsic IsAnisotropic( q::DiagQuadFrm ) -> BoolElt
{
Checks if a quadratic form q is anisotropic over its base ring. 
}
    return not IsIsotropic(q);
end intrinsic;

intrinsic IsIsotropic( V::ModTupFld ) -> BoolElt
{ " } //"
    return not IsIsotropic(V);
end intrinsic;



//////////////////////////////////////////////////
// IsHyperbolic

intrinsic IsHyperbolic( q::DiagQuadFrm ) -> BoolElt
{
Checks if a quadratic form q is isotropic over its base ring. 
Currently supported base rings: finite fields, number fields, reals, complex numbers.
}
    K := q`base_ring;
    require IsField(K): "The base ring of K must be a field.";
    //require ISA(Category(K), FldFin) or ISA(Category(K), FldNum): "The base field is not supported!";
    
    if ISA(Category(K), FldRe) then
        return is_hyperbolic_RR(q);
    elif ISA(Category(K), FldCom) then
        return is_hyperbolic_CC(q);
    elif ISA(Category(K), FldRat) then
        return is_hyperbolic_QQ(q);
    elif ISA(Category(K), FldFin) then
        return is_hyperbolic_FF(q);
    elif ISA(Category(K), FldNum) then
        return is_hyperbolic_NF(q);
    elif ISA(Category(K), FldFun) then
        if IsGlobal(K) then
            return is_hyperbolic_GFF(q);
        end if;
    elif ISA(Category(K), FldFunRat) then
        if IsGlobal(K) then
            return is_hyperbolic_RFF(q);
        end if;
    else
        require false: "The base field is not supported!";
    end if;
    
    return false;
end intrinsic;

intrinsic IsHyperbolic( V::ModTupFld ) -> BoolElt
{ " } //"
    require IsPolarSpace(V): "Argument must be a quadratic space!";
    require PolarSpaceType(V) eq "quadratic space": "Argument must be a quadratic space!";
    K := BaseField(V);
    require Characteristic(K) ne 2: "Characteristic 2 is not supported";
    
    q := Diagonalization(V);
    return IsHyperbolic(q);
end intrinsic;


//////////////////////////////////////////////////
// AnisotropicDimension and WittIndex

intrinsic AnisotropicDimension( q::DiagQuadFrm ) -> RngIntElt
{
Every non-degenerate quadratic form q decomposes (uniquely up to an isometry) 
as q ~ q_1 + ... + c*(1, -1), where q_1 is anisotropic. 
The form q_1 is called an anisotropic part of q.

Given a quadratic form q over a field K, 
this function returns a dimension of an anisotropic part of q.
}
    K := q`base_ring;
    require IsField(K): "The base ring of K must be a field.";
    //require ISA(Category(K), FldFin) or ISA(Category(K), FldNum): "The base field is not supported!";
    
    if ISA(Category(K), FldRe) then
        return anisotropic_dimension_RR(q);
    elif ISA(Category(K), FldCom) then
        return anisotropic_dimension_CC(q);
    elif ISA(Category(K), FldRat) then
        return anisotropic_dimension_QQ(q);
   elif ISA(Category(K), FldFin) then
       return anisotropic_dimension_FF(q);
    elif ISA(Category(K), FldNum) then
        return anisotropic_dimension_NF(q);
    elif ISA(Category(K), FldFun) then
        if IsGlobal(K) then
            return anisotropic_dimension_GFF(q);
        end if;
    elif ISA(Category(K), FldFunRat) then
        if IsGlobal(K) then
            return anisotropic_dimension_RFF(q);
        end if;
    else
        require false: "The base field is not supported!";
    end if;
    
    return false;
end intrinsic;

intrinsic AnisotropicDimension( V::ModTupFld ) -> RngIntElt
{ " } //"
    require IsPolarSpace(V): "Argument must be a quadratic space!";
    require PolarSpaceType(V) eq "quadratic space": "Argument must be a quadratic space!";
    K := BaseField(V);
    require Characteristic(K) ne 2: "Characteristic 2 is not supported";
    
    q := Diagonalization(V);
    return AnisotropicDimension(q);
end intrinsic;

intrinsic WittIndex( q::DiagQuadFrm ) -> RngIntElt
{
The Witt index of a nondegenerate quadratic form q.
}
    require IsNondegenerate(q): "The form must be non-degenerate!";
    return Integers() ! (( Dimension(q) - AnisotropicDimension(q) )/2);
end intrinsic;

intrinsic WittIndex( V::ModTupFld ) -> RngIntElt
{ " } //"
    require IsPolarSpace(V): "Argument must be a quadratic space!";
    require PolarSpaceType(V) eq "quadratic space": "Argument must be a quadratic space!";
    K := BaseField(V);
    require Characteristic(K) ne 2: "Characteristic 2 is not supported";
    
    q := Diagonalization(V);
    return WittIndex(q);
end intrinsic;

//////////////////////////////////////////////////
// AreIsometric

intrinsic AreIsometric( q1::DiagQuadFrm, q2::DiagQuadFrm ) -> BoolElt
{
Gived two quadratic forms, this function checks if they are isomteric.
}
    if Dimension(q1) ne Dimension(q2) then
        return false;
    end if;
    return IsHyperbolic( q1 - q2 );
end intrinsic;

intrinsic AreIsometric( q1::DiagQuadFrm, V2::ModTupFld ) -> BoolElt
{ " } //"
    return AreIsometric(q1, Diagonalization(V2));
end intrinsic;

intrinsic AreIsometric( V1::ModTupFld, q2::DiagQuadFrm ) -> BoolElt
{ " } //"
    return AreIsometric(Diagonalization(V1), q2);
end intrinsic;

intrinsic AreIsometric( V1::ModTupFld, V2::ModTupFld ) -> BoolElt
{ " } //"
    return AreIsometric(Diagonalization(V1), Diagonalization(V2));
end intrinsic;


//////////////////////////////////////////////////
// AreSimilar

intrinsic AreSimilar( q1::DiagQuadFrm, q2::DiagQuadFrm ) -> BoolElt
{
Gived two quadratic forms, this function checks if they are similar.
}
    return IsHyperbolic( q1 - q2 );
end intrinsic;

intrinsic AreSimilar( q1::DiagQuadFrm, V2::ModTupFld ) -> BoolElt
{ " } //"
    return AreSimilar(q1, Diagonalization(V2));
end intrinsic;

intrinsic AreSimilar( V1::ModTupFld, q2::DiagQuadFrm ) -> BoolElt
{ " } //"
    return AreSimilar(Diagonalization(V1), q2);
end intrinsic;

intrinsic AreSimilar( V1::ModTupFld, V2::ModTupFld ) -> BoolElt
{ " } //"
    return AreSimilar(Diagonalization(V1), Diagonalization(V2));
end intrinsic;



