freeze;

import "quadratic_forms.m" : primes_that_matter;
import "../helpers/varia.m" : ZZ, FF2;


function rank2( G )
// This function computes the 2-rank of an abelian group G
    
    I := AbelianInvariants(G);
    return #[n : n in I | IsEven(n)];
end function;

function S_singular_coords(a, GB, SS)
// for a basis of EE_S of a form B\cup L
// where B is a basis of EE compatible with GB
// and L is a basis associated to SS
// this function returns coordinates of an element a

    // coordinates with respect to L/SS-part of the basis
    e := [ZZ| Valuation(a,p) mod 2 : p in SS ];
    // coordinates w.r.t B/GB part of the basis
    s := [ZZ| (1 - LegendreSymbol(a, p))/2 : p in GB ];
    return s cat e;
end function;

function singular_coords(a, GB)
// this function computes coordinates of a in basis
// compatible with a set PP
    
    return [ZZ| (1 - LegendreSymbol(a,p))/2 : p in GB ];
end function;

function point_coords(p, B)
// This function computes the coordinates of [p] + 2Pic X
// w.r.t the basis compatible with B. 
// Here B is a basis of the group of singular elements
    
    return [ZZ| (1 - LegendreSymbol(b,p))/2 : b in B ];
end function;


function simplify( f )
// Given an element in product representation,
// this function gets rid of all squares

    C, E := ProductRepresentation(f);
    return ProductRepresentation( C, [ IsOdd(e) select 1 else 0 : e in E ] );
end function;


function build_point_basis( S ) 
// This function constructs:
//  - GB a basis of the group PicX/2PicX
//  - GB_  set of points s.t. GB cat GB_ generate PicX
//  - GB cat GB_ are disjoint with S

    K := FunctionField(Universe(S));
    // Picard group, psi: Div->Pic, phi:Pic->Div
    PicX, phi, psi := ClassGroup(K);
    PPicX, kappa := PicX/(2*PicX);
    n := rank2(PicX);
    N := #AbelianInvariants(PicX);
    // successively build the basis
    GB  := [];
    GB_ := [];

    subPPicX   := sub< PPicX | >;
    subPicX  := sub< PicX |  >;
    // brute force search
    places := PlaceEnumInit(K);
    // first build GB
    while (subPPicX ne PPicX) or (subPicX ne PicX) do
        // take a next point not in S
        p := PlaceEnumNext(places);
        while p in S do
            p := PlaceEnumNext(places);
        end while;
        // take its classes in PicX and PPicX
        p_ := psi(1*p);
        p__ := kappa(p_);
        // check if it contributes
        if not p__ in subPPicX then
            Append(~GB, p);
            subPPicX := sub< PPicX | subPPicX, p__ >;
            subPicX := sub< PicX | subPicX, p_ >;
        elif not p_ in subPicX then
            Append(~GB_, p);
            subPicX := sub< PicX | subPicX, p_ >;
        end if;
    end while;

    return GB, GB_;
end function;

function build_compatible_basis( GB )
// Given a basis GB of PicX/2PicX, this function constructs
// a compatible basis B of EE

    // Picard group, psi: Div->Pic, phi:Pic->Div
    K := FunctionField(Universe(GB));
//    K := FunctionField( GB[1] );
    PicX, phi, psi := ClassGroup(K);

    // first find *any* basis of EE
    F := ConstantField(K);
    zeta := K!Nonsquare(F);
    // start with a constant nonsquare
    B_ := [ zeta ];
    // loop over genarators of the Picard group of even finite orders 
    // (notice that Magma returns 0 for infinite order)
    for g in Generators(PicX) do
        o := Order(g);
        if IsOdd(o) or (o eq 0) then 
            continue; 
        end if;
        // o*g == 0 in the Picard group, hence the corresponding divisor 
        // is principal ==> find its (unique) generator and attach it to
        // the basis B_
        b := Basis(o*phi(g))[1];
        Append(~B_, simplify(b));
    end for;
    
    // now find *the compatible* basis
    n := #B_;
    // find the change-of-basis matrix
    M := Matrix(FF2, [ singular_coords(b, GB) : b in B_ ]);
    N := M^-1;
    // construct the sought basis compatible with PP
    B := [ K | &* [ B_[j]^(ZZ!N[k][j]) : j in [1..n] ] : k in [1..n] ];
    
    return B;
end function;

function lambda_of_p(p, GB, GB_, B, BU)
// input:
//  - p place of a global function field K
//  - GB sequence of places forming a basis of PicX/2PicX
//  - GB_ sequence of places s.t. GB cat GB_ span PicX
//    (p is NOT in GB cat GB_)
//  - B basis of the group of singular elements compatible with GB
//  - BU sequence of fixed uniformizers for GB
// output: lambda s.t.
//  - ord_p(lambda) is odd
//  - lambda is {p} \cup BB singular
//  - if ord_b(lambda) is even, then lambda is a local 
//    square at b
//  - if ord_b(lambda) is odd, then lambda is in the local
//    square class of the corresponding uniformizer
    
    K := FunctionField(p);
    // Picard group, psi: Div->Pic, phi:Pic->Div
    PicX, phi, psi := ClassGroup(K);
    
    // find the coordinates of p in PicX/2PicX  w.r.t. GB
    E := point_coords(p, B);
    // ZZ lattice
    m := #GB;
    n := #GB_;
    U := FreeAbelianGroup(1 + m + n);
    // sublattice
    V := Kernel(h) where
        h := hom< U->PicX | [psi(1*p)] cat [psi((2-E[j])*GB[j]) : j in [1..m]] cat [psi(2*b) : b in GB_]>;
    // find a vector v in V st v[1] is odd 
    // it must be one of generators since otherwise
    // all linear combinations would have even first
    // coordinate (which we know to be false)
    U_ := StandardLattice(1 + m + n);
    V_ := sub< U_ | [ U_!(Eltseq(U!v)) : v in Generators(V) ]>;
    v := [0];
    for v_ in Basis(V_) do
        if IsOdd((U_!v_)[1]) then
            v := U_!v_;
            break;
        end if;
    end for;
    assert IsOdd(v[1]);

    // find a (unique) generator of the Riemann-Roch space
    // corresponding to v
    la := Basis(D)[1] where 
        D := v[1]*p + &+ [ DivX | (2-E[j])*v[1+j]*GB[j] : j in [1..m] ] 
        + &+ [ DivX | 2*v[1+m+j]*GB_[j] : j in [1..n] ] where
        DivX := DivisorGroup(K);
    // make sure lambda satisfies all local square conditions
    la *:= ProductRepresentation([ b : b in B_ ], [ 1 : b in B_ ]) where
        B_ := [K!1] cat [ B[j] : j in [1..#B] | (LegendreSymbol(la, GB[j]) eq -1) or (LegendreSymbol(la*BU[j], GB[j]) eq -1) ];
    
    return K ! simplify(la);
end function;

function GFF_anisotropic_part_dim3( q )
// Given a quadratic form q = <a1 , . . . , an> of anisotropic dimension 3, 
// this function constructs an element α ∈ K, such that dim_a( q ⊥ <−α>) = 2.
// This is an implementation of [D-MKR202?]

// TODO: this should be converted to an approximation method similar to number fields.

    // base field
    K  := BaseField(q);
    // find the primes that matter
    PP  := primes_that_matter(q);
    // basis of PicX/2PicX dijoint with S
    GB, GB_ := build_point_basis( PP );
    vprint DQF_verbosity : "GB:", GB;
    vprint DQF_verbosity : "GB':", GB_;
    // fix some uniformizers
    unifs := [ UniformizingElement(b) : b in GB ];
    // basis of EE compatible with GB
    B := build_compatible_basis( GB );
    // expand it to a basis of EE_S
    // find lambdas for every place in S
    LA := [ K | lambda_of_p(p, GB, GB_, B, unifs) : p in PP cat GB_ ];
    
    J := [ j : j in [1..#PP] | AnisotropicDimension(q, PP[j]) eq 3 ];
    vprint DQF_verbosity :  "J =", J;
    V := VectorSpace(FF2, #B + #LA);
    Vset := Set(V);

    // loop over these places
    for j in J do
        // for each place find the corresponding *alternative*
        // of congruences
        p := PP[j];
        la := LA[j];
        
        M := Matrix(FF2, [
            [ (LegendreSymbol(a,p) eq -1) select 1 else 0 : a in A ],
            [ Valuation(a,p) : a in A ]
        ]) where A := B cat LA;
        // the discriminant of q will drop the dimension
        W := Vector(FF2, [
            (LegendreSymbol(d,p) + LegendreSymbol(d*la,p) eq 1) select 0 else 1,
            Valuation(d, p)
        ]) where d := Discriminant(q);

        // find the *set* solutions
        v0 := Solution( Transpose(M), W );
        V0 := NullspaceOfTranspose(M);
        Wset := { v0 + v : v in V0 };
        Vset := Vset diff Wset;
    end for;
    // convert a solution to an element of K
    A := B cat LA;
    v := Rep(Vset);
    vprint DQF_verbosity :  "v =", v;
    alpha := ProductRepresentation(
            [ A[i]^(Integers() ! v[i]) : i in [1..#A] ], 
            [1 : i in [1..#A]]
            );
    alpha := simplify(alpha);
    vprint DQF_verbosity :  "alpha =", alpha;


    return alpha;
end function;

function GFF_anisotropic_part_dim2( q )
// Given a quadratic form q = <a1 , . . . , an> of anisotropic dimension 2, 
// this function constructs ist anisotropic part <a,b>.
// This is an implementation of [D-MKR202?, Algorithm 2]

    // base field
    K  := BaseField(q);
    // find the primes that matter
    S  := primes_that_matter(q);
    
    // make sure the Witt index is even
    if IsOdd(WittIndex(q)) then
        q := q + HyperbolicForm(K, 2);
        q`primes := S;
    end if;
    
    d := Discriminant(q);
    
    // basis of PicX/2PicX dijoint with S
    GB, GB_ := build_point_basis( S );
    vprint DQF_verbosity : "GB:", GB;
    vprint DQF_verbosity : "GB':", GB_;
    // fix some uniformizers
    unifs := [ UniformizingElement(b) : b in GB ];
    // basis of EE compatible with GB
    B := build_compatible_basis( GB );
    // expand it to a basis of EE_S
    // find lambdas for every place in S
    LA := [ K | lambda_of_p(p, GB, GB_, B, unifs) : p in S cat GB_ ];
    S := GB cat GB_ cat S;
    
    found := false;
    // initialization of places' search
    places := PlaceEnumInit(K);
    while not found do
        B_ := B cat LA;
        vprint DQF_verbosity : "basis:", B_;
    
        // vector of exponents of Hasse invariants
        W := Vector(FF2, [ZZ | (1 - HasseMinkowskiInvariant(q, P))/2 : P in S ]);
        vprint DQF_verbosity : "W =", W;
    
        // matrix of Hilbert symbols
        H := Matrix(FF2, [[ZZ | (1 - HilbertSymbol(b, d, P))/2  : b in B_ ] : P in S ]);
        vprint DQF_verbosity : "H =", H;
        
        // V solution to H*V == W
        found, V, _ := IsConsistent(Transpose(H), W);
        
        if found then
            vprint DQF_verbosity : "V =", V;
            // the sought element
            alpha := ProductRepresentation(B_, V_) where V_ := Eltseq( ChangeRing(V,ZZ));
            alpha := simplify(alpha);
        else
            P := PlaceEnumNext(places);
            while P in S do
                P := PlaceEnumNext(places);
            end while;
            lambda := lambda_of_p(P, GB, GB_, B, unifs);
            Append(~S, P);
            Append(~LA, lambda);
            vprint DQF_verbosity : "Expanding S. Added P =", P;
        end if;
        
    end while;
    
    return DiagonalQuadraticForm(K, [alpha, -alpha*d]);
end function;


function anisotropic_part_GFF( q )
// Given a diagonal quadratic form q over a global function field, this function computes its 
// anisotropic part q_a, that is an anisotropic form s.t. q = q_a ⊥ w*<1,-1>.

    K := BaseField(q);
    P := primes_that_matter(q);
    d := AnisotropicDimension(q);
    
    // first short-cut
    if d eq Dimension(q) then
        return q;
    end if;
    
    // initialize the anisotropic part to be an empty form
    qa := DiagonalQuadraticForm(K, []);
    
    // second short-cut
    C := Coefficients(q);
    for c in C do
        q_ := q - DiagonalQuadraticForm(K, [c]);
        // substracting an existing coeff will not change primes
        // don't waste time on a new factorization
        q_`primes := P;
        if AnisotropicDimension(q_) lt d then
            q := q_;
            q`primes := P;
            qa +:= DiagonalQuadraticForm(K, [c]);
            d -:= 1;
        end if;
    end for;
    
    if d eq 4 then
        a_  := DiagonalQuadraticForm(K, [ One(K) ]);
        qa +:= a_;
        q  -:= a_;
        d   := 3;
        assert2 AnisotropicDimension(q) eq 3;
    end if;
    
    if d eq 3 then
        a   := GFF_anisotropic_part_dim3( q );
        a_  := DiagonalQuadraticForm(K, [a]);
        qa +:= a_;
        q  -:= a_;
        d   := 2;
        assert2 AnisotropicDimension(q) eq 2;
    end if;

    if d eq 2 then
        q_ := GFF_anisotropic_part_dim2( q );
        qa +:= q_;
    end if;
    
    if d eq 1 then
        qa +:= DiagonalQuadraticForm(K, [Discriminant(q)]);
    end if;
    
    return qa;
end function;
