freeze;

declare attributes FldFun: PythNum, PythNumElt, Hght;

//////////////////////////////////////////////////
// Level
// (just for completeness)
intrinsic Level( K::FldFun ) -> RngIntElt
{    
This function comutes the level s(K) of a number field K,
i.e. a lowest number s such that -1 is a sum of s squares
it returns 0 if K is formally real.
}
    return LengthOfSOS( K!-1);
end intrinsic;


//////////////////////////////////////////////////
// IsFormallyReal
// just for completeness
intrinsic IsFormallyReal( K::FldFun ) -> BoolElt
{
Returns true iff K is formally real (i.e. -1 is not a sum of squares)
}
    return IsSOS(K!-1);
end intrinsic;

intrinsic IsReal( K::FldFun ) -> BoolElt
{ " } //"
    return IsFormallyReal(K);
end intrinsic;


//////////////////////////////////////////////////
// PythagorasNumber & PythagorasElement

intrinsic PythagorasNumber( K::FldFun ) -> RngIntElt
{
This function computes the Pythagoras number P(K) of a number field K,
i.e. the lowest number p such that every sum of squares in K
is a sum of p squares.
}
    require ISA(Category(ConstantField(K)), FldFin) : "Field of constants must be finite!";
    require IsOdd(Characteristic(K)): "Not implemented in even characteristic!";
    if assigned K`PythNum then
        return K`PythNum;
    end if;
    
    // there are no dyadic primes, hence
    // P(K) must be either 2 or 3, depending on s(K)
    K`PythNum := IsSquare(K!-1) select 2 else 3;
    return K`PythNum;
end intrinsic;

intrinsic PythagorasElement( K::FldFun ) -> FldFunElt
{
Returns an element a of K whose length equals the Pythagoras 
number P(K) of K.
}
    require ISA(Category(ConstantField(K)), FldFin) : "Field of constants must be finite!";
    require IsOdd(Characteristic(K)): "Not implemented in even characteristic!";
    if assigned K`PythNumElt then
        return K`PythNumElt;
    end if;
    
    PK := PythagorasNumber(K);
    sK := Level(K);
    if sK eq 1 then
        // PK == 2
        // find any non-square
        // we are in a global case, hence there is an odd degree place
        // so any nonsquare in the field of constants is fine
        F, m := ExactConstantField(K);
        K`PythNumElt := m(Nonsquare(F));
        assert2 not IsSquare(K`PythNumElt);
    else
        // PK == 3
        // find a place where -1 is not a square
        // it must have an odd degree 
        S := Support( DivisorOfDegreeOne(K) );
        for p in S do
            if IsOdd(Degree(p)) then
                K`PythNumElt := LocalUniformizer(p);
                break;
            end if;
        end for;
    end if;
    return K`PythNumElt;
end intrinsic;

//////////////////////////////////////////////////
// AreWittEquivalent

intrinsic AreWittEquivalent( K::FldFun, L::FldFun ) -> BoolElt
{
Given two fields K and L, this function return true if their Witt rings are isomorphic and false otherwise.
}    
    require ISA(Category(ConstantField(K)), FldFin) : "Not implemented for these types of fields!";
    require ISA(Category(ConstantField(L)), FldFin) : "Not implemented for these types of fields!";

    // Use Szymiczek (1991), Theorem 1.1
    if Characteristic(K) eq 2 and Characteristic(L) ne 2 then
        return false;
    elif Characteristic(K) ne 2 and Characteristic(L) eq 2 then
        return false;
    elif Characteristic(K) eq 2 and Characteristic(L) eq 2 then
        return true;
    else
        // Now both must have odd characteristic
        // Use Szymiczek (1991), Theorem 1.3
        return Level(K) eq Level(L);
    end if;
end intrinsic;

intrinsic AreWittEquivalent( K::FldFun, L::FldFunRat ) -> BoolElt
{ " } //"
    require ISA(Category(ConstantField(K)), FldFin) : "Not implemented for these types of fields!";
    require ISA(Category(ConstantField(L)), FldFin) : "Not implemented for these types of fields!";

    // Use Szymiczek (1991), Theorem 1.1
    if Characteristic(K) eq 2 and Characteristic(L) ne 2 then
        return false;
    elif Characteristic(K) ne 2 and Characteristic(L) eq 2 then
        return false;
    elif Characteristic(K) eq 2 and Characteristic(L) eq 2 then
        return true;
    else
        // Now both must have odd characteristic
        // Use Szymiczek (1991), Theorem 1.3
        return Level(K) eq Level(L);
    end if;
end intrinsic;

intrinsic AreWittEquivalent( K::FldFun, L::FldNum ) -> BoolElt
{ " } //"
    require ISA(Category(ConstantField(K)), FldFin) : "Not implemented for these types of fields!";

    // Use Szymiczek (1991), Theorem 1.3
    return false;
end intrinsic;

intrinsic AreWittEquivalent( K::FldFun, L::FldRat ) -> BoolElt
{ " } //"
    require ISA(Category(ConstantField(K)), FldFin) : "Not implemented for these types of fields!";

    // Use Szymiczek (1991), Theorem 1.3
    return false;
end intrinsic;




