 freeze;
declare attributes FldFun: sos1;    // decomposition of -1 into sum of squares

import "../helpers/varia.m" : ZZ, FF2;


// Correction for NormEquation
intrinsic NormEquation(F::FldFun, a::FldFunGElt) -> BoolElt, []
  {Tries to solve N(x) = a in the global function field F}

  require IsGlobal(F) : "Function field is not global";
  C, mC := ClassGroup(F);
  S := Support(Divisor(F!a));
  d := 0;
  lp := [];
  i := 1;
  U := sub<C|[x @@ mC : x in S]>;
  while U ne C do
    while i gt #lp do
      d +:= 1;
      lp := Places(F, d);
      i := 1;
    end while;
    c := lp[i]@@mC;
    if not c in U then
      U := sub<C|U, c>;
      Include(~S, lp[i]);
    end if;
    i+:= 1;
  end while;
  T := {};
  for s in S do
    if IsFinite(s) then
      m := Minimum(Ideal(s));
      assert IsPrime(m);
      T join:= {Place(x) : x in Decomposition(Order(Ideal(s)), m)};
    else
      T join:= {x : x in InfinitePlaces(F)};
    end if;
  end for;

  SF, mSF := SUnitGroup(T);
  N := hom<SF -> SF | [(F!Norm(SF.i @ mSF))@@mSF : i in [1..Ngens(SF)]]>;
  r := (F!a)@@mSF;
  if r in Image(N) then
    return true, [(r@@N)@mSF];
  else
    return false, _;
  end if;
end intrinsic;


//////////////////////////////////////////////////
// Helper functions for SolveSOS

function ssos_len2( a )
// Given an element of length 2, this function returns c1, c2
// such that c1^2 + c2^2 = a
    K := Parent(a);
    
    test, i := IsSquare(K!-1);
    if test then
        c1 := (a+1)/2;
        c2 := i*(a-1)/2;
    else
        P := PolynomialRing(K);
        L := ext< K | P.1^2 + 1 >;
        test, v := NormEquation(L, a);
        assert test;
        // convert to coeffs
        v := Eltseq( v[1] );
        c1 := v[1];
        c2 := v[2];
    end if;
    
    assert c1^2 + c2^2 eq a;
    return [ c1, c2 ];
end function;

function ssos_len3( a )
// Given an element of length 3, this function returns c1, c2, c3
// such that c1^2 + c2^2 + c3^2 = a
    K := Parent(a);
    K_ := ConstantField(K);
    assert Level(K_) eq 2;

/*    
    if assigned K`sos1 then
        d1 := K`sos1[1];
        d2 := K`sos1[2];
    else
        P := PolynomialRing(K);
        L := ext< K | P.1^2 + 1 >;
        test, v := NormEquation(L, a : Nice := true);
        assert test;
        K`sos1 := Eltseq(v[1]);
        d1 := K`sos1[1];
        d2 := K`sos1[2];   
    end if;
*/
    P := PolynomialRing(K_);
    L_ := ext< K_ | P.1^2 + 1 >;
    test, v := NormEquation(L_, K_!-1);
    assert test;
    v := Eltseq(v);
    d1 := v[1];
    d2 := v[2];   

    c := [ (a+1)/2, d1*(a-1)/2, d2*(a-1)/2 ];

    assert c[1]^2 + c[2]^2 + c[3]^2 eq a;
    return c;
end function;

intrinsic SolveSumOfSquares( a::FldFunElt ) -> []
{
Let a be a sum of squares in K. Assume that the length of a is n.
This function solves the equation a = x1^2 + ... + xn^2 in K.
}
    K := Parent(a);
    require IsGlobal(K) : "The base field mus t be global!";
    require IsSumOfSquares(a) : "The argument must be a sum of squares!";
    
    test, sqrt := IsSquare(a);
    if test then
        return [ sqrt ];
    end if;
    
    l := LengthOfSOS(a);
    if l eq 2 then 
        return ssos_len2(a);
    else 
        return ssos_len3(a);
    end if;
end intrinsic;

intrinsic SolveSOS( a::FldFunElt ) -> []
{ " } //"
    return SolveSumOfSquares(a);
end intrinsic;


//////////////////////////////////////////////////
// IsSumOfSquares
// (just for completeness)
intrinsic IsSumOfSquares( a::FldFunElt ) -> BoolElt
{ " }
    K := Parent(a);
    require IsGlobal(K) : "The base field mus t be global!";
    // require ISA(Category(ConstantField(K)), FldFin) : "Field of constants must be finite!";
    return true;
end intrinsic;

intrinsic IsSOS( a::FldFunElt ) -> BoolElt
{ " }
    return IsSumOfSquares(a);
end intrinsic;


//////////////////////////////////////////////////
// LengthOfSOS

function length_nondyadic( a, P )
// Given a nonzero element a of a global function field K and a place P, 
// this fuction determines the minimal number of summands needed to express 
// a as a sum of squares in a completion K_P.  
    
    // the residue field
    Kp := ResidueClassField(P);
    // and valuation
    v := Valuation(a,P);
    // two cases to consider
    if IsOdd(v) then
        // possible values are 2 and 3
        q := DiagonalQuadraticForm(Kp, [Kp!1, Kp!1]);
        return IsIsotropic(q) select 2 else 3;
    else
        // possible values are 1 and 2
        return IsLocalSquare(a, P) select 1 else 2;
    end if;
end function;

intrinsic LengthOfSOS( a::FldFunElt ) -> RngIntElt
{
Given a nonzero element a of a global function field K, this function determines
minimal number of summands needed to express a as a sum of squares. 
It returns 0 if a is not a sum of squares
}
    require not IsZero(a): "The argument must be nonzero!";
    K := Parent(a);
    require IsGlobal(K) : "The base field mus t be global!";
    // start with trivial cases
    if not IsSOS(a) then
        return 0;
    end if;
    if IsSquare(a) then
        return 1;
    end if;
    // the possible values now are 2, 3, 4
    // find all the primes of K that matter for a
    primes := Support(Divisor(a));
    vprint DQF_verbosity : "Primes to be tested:", primes;
    if IsEmpty(primes) then
        // a is a constant
        return LengthOfSOS( ConstantField(K)!a );
    end if;

    n := Max([ length_nondyadic(a, P) : P in primes ]);
    return n;
end intrinsic;

intrinsic LengthOfSumOfSquares( a::FldFunElt ) -> RngIntElt
{ " } //"
    return LengthOfSOS(a);
end intrinsic;


