freeze;

QQ := Rationals();
ZZ := Integers();

//////////////////////////////////////////////////
// RealRootBound

function bound_Hong( f )
    // Hong bound for real roots of f
    if LeadingCoefficient(f) lt 0 then
        f := -f;
    end if;
    L := Coefficients(f);
    // polys with only positive coeffs have no positive roots
    if forall{ a : a in L | a ge 0 } then
        return 0;
    end if;
    n := Degree(f);
    H := Max([ 
        Min([ (-L[j+1]/L[i+1])^(1/(i-j)) : i in [j+1..n] | L[i+1] gt 0 ]) 
        : j in [0..n-1] | L[j+1] lt 0 
    ]);
    return 2*H;
end function;

function bound_local_max( f )
    // Vigklas linear local-max bound for real roots of f
    if LeadingCoefficient(f) lt 0 then
        f := -f;
    end if;
    L := Coefficients(f);
    n := Degree(f);
    // index initialization
    i := n;
    // exponent initialization
    t := 1;
    // upper bound initialization
    M := 0;
    for j in [n-1..0 by -1] do
        if L[j+1] gt L[i+1] then
            i := j;
            t := 1;
        end if;
        if L[j+1] lt 0 then
            M := Max([ M, (-2^t*L[j+1]/L[i+1])^(1/(i-j)) ]);
            t +:= 1;
        end if;
    end for;
    return M;
end function;


intrinsic RealRootBound( f::RngUPolElt : algorithm := "Hong" ) -> FldReElt
{ 
    Returns an upper bound for real roots of a polynomial f. 
    Possible values of the optional parameter algorithm are:
        - "Hong" - use Hong bound (default)
        - "local max" - use Vigklas linear time local-max bound
}
    if algorithm eq "local max" then
        return bound_local_max(f);
    end if; 
    return bound_Hong(f);
end intrinsic;

//////////////////////////////////////////////////
// RealRootIntervals


function sign_changes( L )
    // Countes the number of sign changes in a sequence
    L := [ Sign(l) : l in L | l ne 0 ];
    M := [ L[j] : j in [1..#L-1] | L[j] ne L[j+1] ];
    return #M;
end function;

function AlesinaGaluzzi( f, a, b )
    // Given a polynomial f, this function isolates its roots in (a,b)
    // using Alesina-Galuzzi algorithm
    
    
    if b eq Infinity() then
        b := RealRootBound(f);
        // round to an integer
        b := Ceiling(b);
    end if;
    
    if a ge b then
        return [];
    end if;
    
    n := Degree(f);
    P := Parent(f);
    x := P.1;
    
    g := P!( (1+x)^n*Evaluate(f, (a+b*x)/(1+x) ) );
    v := sign_changes( Coefficients(g) );
    
    if v eq 0 then
        return [];
    elif v eq 1 then
        return [ <a,b> ];
    else
        c := (a+b)/2;
        L := AlesinaGaluzzi(f, a, c);
        U := AlesinaGaluzzi(f, c, b);
        if Evaluate(f, c) eq 0 then
            return L cat [ <c,c> ] cat U;
        else
            return L cat U;
        end if;
    end if;
end function;

intrinsic RealRootIntervals( f::RngUPolElt ) -> []
{
    Given a polynomial f, this function construct a list of pair of 
    closed intervals [a,b] that isolate real roots of f. The endpoints 
    a nad b are rational numbers with denominators powers of 2.
}
    K := CoefficientRing(f);
    P := Parent(f);
    if not IsField(K) then
        P := ChangeRing(P, FieldOfFractions(K));
        f := P!f;
    end if;
    x := P.1;
    
    f := SquarefreePart(f);
    // isolate positive roots
    A := AlesinaGaluzzi(f, 0, Infinity());
    // isolate negative roots
    B := AlesinaGaluzzi( Evaluate(f, -x), 0, Infinity());
    Reverse(~B);
    B := [ <-b[2],-b[1]> : b in B ];
    // isolate zero
    if Evaluate(f, 0) eq 0 then
        return B cat [ <0,0> ] cat A;
    else 
        return B cat A;
    end if;
end intrinsic;

