freeze;

import "quadratic_forms.m" : signatures_builtin, primes_that_matter;
import "anisotropic_part.m" : strong_ordering_separation_v3;
import "../helpers/varia.m" : ZZ, FF2;
import "varia.m" : initialize_prime_iterator, next_prime, positive_approximation;

forward isotropic_vector_NF;

/////////////////////////////////////////////
//
// Helper function

function relative_pseudo_class_group( L )
// given a field extension L/K compute the generators of the
// pseudo-class group Cl_i(L/K). Return two sets:
// S places in K and S_ places in L

    // retrieve K
    K := BaseField(L);
    L_ := AbsoluteField(L);
    OL := MaximalOrder(L);
    OL_ := MaximalOrder(L_);

    // class group Cl(L)
    // mClL : Cl(L) -> Ideals(L)
    ClL, mClL := ClassGroup(L_);
    ClK, mClK := ClassGroup(K);
    
    // compute the image of ClK in ClL
    iClK := sub< ClL | [ ( &+ [ a*OL_ : a in Generators(mClK(Zero(ClK))) ] ) @@ mClL 
                : g in Generators(ClK) ] >;
//    iClK := sub< ClL | [ ClLm( DivL ! mClK(g) ) : g in Generators(ClK) ] >;
    // now we have the ralative pseudo-class group
    Cli, kappa := ClL/iClK;
    
    // find a set of places that generate Cli
    S_ := [ mClL(g @@ kappa) : g in Generators(Cli) ];
    S_ := &join [ Support(g) : g in S_ ];
    assert2 sub< Cli | [ kappa( p @@ mClL ) : p in S_ ]> eq Cli; 
    // find the places in K dominated by S_
    S := &join [ Support(Norm( &+ [ a*OL : a in Generators(p)] )) : p in S_ ];
    //S := [ Support(Norm(p))[1] : p in S_ ];
    
    return S, S_;
end function;


/////////////////////////////////////////////
//
// Isotropic vectors for various dimes

function isovec2_NF( q )
// given a quadratic form of dimension 2, return its isotropic vector
    assert Dimension(q) eq 2;
    
    C := Coefficients(q);
    test, a := IsSquare( -C[1]*C[2] );
    assert test;
    K := BaseRing(q);
    w := Vector(K, [ C[2], a ]);
    assert IsZero(Evaluate(q, w));
    
    return w;
end function;

function isovec3_NF_builtin( q )
// given a quadratic form of dimension 3, return its isotropic vector
// this function uses built-in function NormEquation
    assert Dimension(q) eq 3;
    
    K := BaseRing(q);
    C := Coefficients(q);
    
    // check if there is a dimension 2 isotropic subform
    test, a := IsSquare( -C[1]*C[2] );
    if test then return Vector(K, [ C[2], a, 0 ]); end if;
    test, a := IsSquare( -C[1]*C[3] );
    if test then return Vector(K, [ C[3], 0, a ]); end if;
    test, a := IsSquare( -C[2]*C[3] );
    if test then return Vector(K, [ 0, C[3], a ]); end if;
    
    a := C[1]; b := C[2]; c := C[3];
    x := PolynomialRing(K).1;
    L := ext< K | x^2 + b/a >;
    
    test, v := NormEquation(L, -c/a);
    assert test;
    // convert to coeffs
    v := Eltseq( v[1] );
    test, v3 := IsSquare( -a/c*v[1]^2 - b/c*v[2]^2 );
    assert test;

    w := Vector(K, [v[1], v[2], v3]);
    assert IsZero(Evaluate(q, w));
    
    return w;
end function;

function isovec4_NF_v1( q )
// Compute an isotropic vector of a 4-dimensional form.
// This is exactly the version described in my paper, it is much slower than v.2.
// But on the other hand, sometimes it gives much nicer results.
    assert Dimension(q) eq 4;
    
    K := BaseRing(q);
    C := Coefficients(q);
    
    // check if there is a dimension 3 isotropic subform
    q1 := DiagonalQuadraticForm(K, C[ [2,3,4] ]);
    if IsIsotropic(q1) then
        v := Eltseq( isovec3_NF_builtin(q1) );
        return Vector(K, [0, v[1], v[2], v[3]]);
    end if;
    q1 := DiagonalQuadraticForm(K, C[ [1,3,4] ]);
    if IsIsotropic(q1) then
        v := Eltseq( isovec3_NF_builtin(q1) );
        return Vector(K, [v[1], 0, v[2], v[3]]);
    end if;
    q1 := DiagonalQuadraticForm(K, C[ [1,2,4] ]);
    if IsIsotropic(q1) then
        v := Eltseq( isovec3_NF_builtin(q1) );
        return Vector(K, [v[1], v[2], 0, v[3]]);
    end if;
    q1 := DiagonalQuadraticForm(K, C[ [1,2,3] ]);
    if IsIsotropic(q1) then
        v := Eltseq( isovec3_NF_builtin(q1) );
        return Vector(K, [v[1], v[2], v[3], 0]);
    end if;

    // two quadratic extensions
    x := PolynomialRing(K).1;
    L := ext< K | x^2 + C[2]/C[1] >;
    L_ := AbsoluteField(L);
    M := ext< K | x^2 + C[4]/C[3] >;
    M_ := AbsoluteField(M);
    OL := MaximalOrder(L);
    OL_ := MaximalOrder(L_);
    OM := MaximalOrder(M);
    OM_ := MaximalOrder(M_);
    
    // find the big enough set of primes
    S0 := primes_that_matter(q);
    S1, S1_ := relative_pseudo_class_group(L);
    S2, S2_ := relative_pseudo_class_group(M);
    
    // S0-class group
    // mClS0 : ClS0 -> Div(K)
//    ClS0, mClS0 := SClassGroup(Seqset(S0));
//    S3 := &cat [ Support(ClS0m(p)) : p in Generators(ClS0) ];
    
    S := Seqset(S0) join S1 join S2;
    //S := Setseq(S);
    
    // main loop
    found := false;
    // initialization of places' search
    it := initialize_prime_iterator(K);
    while not found do
        // places of L dominating S
        SL_ := &join [ Support(&+ [ g*OL_ : g in Generators(p) ]) : p in S ];
        //SL := &cat [ Support(DivL ! p) : p in S ];
        // places of M dominating S
        SM_ := &join [ Support(&+ [ g*OM_ : g in Generators(p) ]) : p in S ];
        //SM := &cat [ Support(DivM ! p) : p in S ];
    
        // construct the bases of corresponding S*-singular groups
        EE_K, mEE_K := pSelmerGroup(2, S);
        B_K := [ g @@ mEE_K : g in Generators(EE_K) ];
        EE_L, mEE_L := pSelmerGroup(2, SL_);
        B_L := [ g @@ mEE_L : g in Generators(EE_L) ];
        EE_M, mEE_M := pSelmerGroup(2, SM_);
        B_M := [ g @@ mEE_M : g in Generators(EE_M) ];
    
        // coordinates of a_1 in EE_K
        a := Eltseq( mEE_K(C[1]) );
        // coordinates of -a_3 in EE_K
        b := Eltseq( mEE_K(-C[3]) );
        // find the coords of norms of all els in B_L
        A := [ Eltseq( mEE_K(Norm(L!l)) ) : l in B_L ];
        // find the coords of norms of all els in B_M
        B := [ Eltseq( mEE_K(Norm(M!m)) ) : m in B_M ];
        
        // form system (♣)
        AB := Matrix(FF2, A cat B);
        v := Vector(FF2, a) + Vector(FF2, b);
        found, V, _ := IsConsistent(AB, v);
        if found then
            V := Eltseq(ChangeRing(V,ZZ));
            v1v2 := L!ProductRepresentation(B_L, V[1..#B_L]);
            v3v4 := M!ProductRepresentation(B_M, V[#B_L+1..#V]);
        
            test, c := IsSquare( C[1]*Norm( v1v2 ) / (-C[3]*Norm( v3v4 )) ); 
            assert test;

            v := Vector(K, Eltseq(v1v2) cat Eltseq(c*v3v4));
        else
            it, P := next_prime(it);
            while P in S do
                it, P := next_prime(it);
                //P := Ideal(Decomposition(K, RandomPrime(20))[1][1]);
            end while;
            Include(~S, P);
            vprint DQF_verbosity : "Expanding S. Added P =", P;
        end if;
    end while;

    assert IsZero(Evaluate(q, v));
    return Vector(K, v);
end function;

function isovec4_NF_v2( q )
// This version of the function is much faster, but the results may be not so nice
    assert Dimension(q) eq 4;
    
    K := BaseField(q);
    //OK := MaximalOrder(K);
    C := Coefficients(q);
    
    // check if there is a dimension 3 isotropic subform
    q1 := DiagonalQuadraticForm(K, C[ [2,3,4] ]);
    if IsIsotropic(q1) then
        v := Eltseq( isovec3_NF_builtin(q1) );
        return Vector(K, [0, v[1], v[2], v[3]]);
    end if;
    q1 := DiagonalQuadraticForm(K, C[ [1,3,4] ]);
    if IsIsotropic(q1) then
        v := Eltseq( isovec3_NF_builtin(q1) );
        return Vector(K, [v[1], 0, v[2], v[3]]);
    end if;
    q1 := DiagonalQuadraticForm(K, C[ [1,2,4] ]);
    if IsIsotropic(q1) then
        v := Eltseq( isovec3_NF_builtin(q1) );
        return Vector(K, [v[1], v[2], 0, v[3]]);
    end if;
    q1 := DiagonalQuadraticForm(K, C[ [1,2,3] ]);
    if IsIsotropic(q1) then
        v := Eltseq( isovec3_NF_builtin(q1) );
        return Vector(K, [v[1], v[2], v[3], 0]);
    end if;

    // primes to be considered
    S := primes_that_matter(q);
    
    // real places of K
    Soo := RealPlaces(K);
    sgn_q := [ RealSigns(c) : c in C ];
    sgn_q1 := [ sgn_q[1][i] + sgn_q[2][i] : i in [1..#Soo] ];
    sgn_q2 := [ sgn_q[3][i] + sgn_q[4][i] : i in [1..#Soo] ];
    
    // three vectors to be constructed
    // 1. signs at the real places, where either q1 or q2 is anisotropic
    inx := [ i : i in [1..#Soo] | (sgn_q1[i] ne 0) or (sgn_q2[i] ne 0) ];
    u := [ FF2 | 
        (sgn_q1[i] eq -2) select 1 else (
        (sgn_q1[i] eq  2) select 0 else (
        (sgn_q2[i] eq -2) select 0 else 1))
    : i in inx ];
    
    // initialization of places' search
    it := initialize_prime_iterator(K);
    // main loop
    found := false;

    while not found do
        // basis of the group of S-singular elements
        // E_S group of singular elements 
        // psi : OK -> E_S
        E_S, psi := pSelmerGroup(2, Set(S) : Nice);
        B_K := [ K!(g @@ psi) : g in Generators(E_S) ];
        vprint DQF_verbosity : "basis:", B_K;
   
        // three vectors to be constructed
        // 2. Hilbert symbols (a_1, a_2)
        v := [ FF2 | HilbertSymbol(C[1], C[2],p) eq -1 select 1 else 0 : p in S ];
        // 3. Hilbert symbols (-a_3,-a_4)
        w := [ FF2 | HilbertSymbol(-C[3], -C[4], p) eq -1 select 1 else 0 : p in S ];
        
        // three matrices to be constructed
        // 1. signs at the real places, where either q1 or q2 is anisotropic
        U := Matrix(FF2, [[ a eq -1 select 1 else 0 : a in RealSigns(kappa) ] : kappa in B_K ]);
        U := Matrix(Rows(Transpose(U))[inx]);
        // 2. Hilbert symbols (-a_1a_2, kappa)
        V := Matrix(FF2, 
            [[ HilbertSymbol(-C[1]*C[2],kappa,p) eq -1 select 1 else 0 : kappa in B_K ] : p in S ]);
        // 3. Hilbert symbols (-a_3a_4, kappa)
        W := Matrix(FF2, 
            [[ HilbertSymbol(-C[3]*C[4],kappa,p) eq -1 select 1 else 0 : kappa in B_K ] : p in S ]);
        
        // combine the matrices 
        // no idea why it doesn't work
        //ABC := VerticalJoin([A,B,C]);
        // using a dirty hack instead
        UVW := VerticalJoin(VerticalJoin(U,V),W);
        // combine the three vectors
        uvw := Vector(FF2, u cat v cat w);
        
        // check if there is a solution
        found, Xi, _ := IsConsistent( Transpose(UVW), uvw);
        
        if found then
            vprint DQF_verbosity : "Xi =", Xi;
            // the sought element
            b := ProductRepresentation(B_K, Xi_) where Xi_ := Eltseq(ChangeRing(Xi,ZZ));
        else
            it, P := next_prime(it);
            while P in S do
                it, P := next_prime(it);
            end while;
            Append(~S, P);
        end if;
    end while;
    
    q1 := DiagonalQuadraticForm(K, [ -b, C[1], C[2] ]);
    q2 := DiagonalQuadraticForm(K, [  b, C[3], C[4] ]);
    w1 := isovec3_NF_builtin( q1 );
    w2 := isovec3_NF_builtin( q2 );
    
    v := [ w1[2]/w1[1], w1[3]/w1[1], w2[2]/w2[1], w2[3]/w2[1] ];
    assert IsZero(Evaluate(q, v));
    return Vector(K, v);
end function;

function isovec5_NF( q : alg := "v2"  )
    assert Dimension(q) eq 5;
    
    K := BaseField(q);
    //OK := MaximalOrder(K);
    C := Coefficients(q);
    
    // check if there is a dimension 4 isotropic subform
    q1 := DiagonalQuadraticForm(K, C[ [2,3,4,5] ]);
    if IsIsotropic(q1) then
        v := Eltseq( isotropic_vector_NF(q1 : alg := alg) );
        return Vector(K, [0, v[1], v[2], v[3], v[4]]);
    end if;
    q1 := DiagonalQuadraticForm(K, C[ [1,3,4,5] ]);
    if IsIsotropic(q1) then
        v := Eltseq( isotropic_vector_NF(q1 : alg := alg) );
        return Vector(K, [v[1], 0, v[2], v[3], v[4]]);
    end if;
    q1 := DiagonalQuadraticForm(K, C[ [1,2,4,5] ]);
    if IsIsotropic(q1) then
        v := Eltseq( isotropic_vector_NF(q1 : alg := alg) );
        return Vector(K, [v[1], v[2], 0, v[3], v[4]]);
    end if;
    q1 := DiagonalQuadraticForm(K, C[ [1,2,3,5] ]);
    if IsIsotropic(q1) then
        v := Eltseq( isotropic_vector_NF(q1 : alg := alg) );
        return Vector(K, [v[1], v[2], v[3], 0, v[4]]);
    end if;
    q1 := DiagonalQuadraticForm(K, C[ [1,2,3,4] ]);
    if IsIsotropic(q1) then
        v := Eltseq( isotropic_vector_NF(q1 : alg := alg) );
        return Vector(K, [v[1], v[2], v[3], v[4], 0]);
    end if;

    // two subforms
    q1 := DiagonalQuadraticForm(K, [C[1], C[2], C[3]]);
    q2 := DiagonalQuadraticForm(K, [C[4], C[5]]);
    
    // primes to be considered
    T := primes_that_matter(q);
    
    // real places of K
    Soo := RealPlaces(K);
    sgn_q := [ RealSigns(c) : c in C ];
    sgn_q1 := [ sgn_q[1][i] + sgn_q[2][i] + sgn_q[3][i] : i in [1..#Soo] ];
    sgn_q2 := [ sgn_q[4][i] + sgn_q[5][i] : i in [1..#Soo] ];
    
    // find the places where at least one of q1, q2 is anisotropic
    inx := [ i : i in [1..#Soo] | (Abs(sgn_q1[i]) eq 3) or (Abs(sgn_q2[i]) eq 2) ];
    alpha := [ FF2 | 
        (sgn_q1[i] eq -3) select 1 else (
        (sgn_q1[i] eq  3) select 0 else (
        (sgn_q2[i] eq -2) select 0 else 1))
    : i in inx ];
    
    // initialization of places' search
    it := initialize_prime_iterator(K);
    // main loop
    found := false;

    while not found do
        // basis of the group of T-singular elements
        // E_T group of singular elements 
        // psi : OK -> E_T
        E_T, psi := pSelmerGroup(2, Set(T) : Nice);
        B_K := [ K!(g @@ psi) : g in Generators(E_T) ];
        vprint DQF_verbosity : "basis:", B_K;
    
        // subset of T of the primes that *really matter*
        S := [ p : p in T | not IsIsotropic(q1, p) or not IsIsotropic(q2, p) ];
    
        // generators of local square-class groups
        // and elements c s.t. q1 represents c and q2 represents -c (at P)
        HH := [];
        cc := [];
        for P in S do
            V, mV := LocalMultiplicativeGroupModSquares( P );
            Append(~HH, [ mV(v) : v in Generators(V) ]);
            
            for v in V do
                ci := mV(v);
                if IsIsotropic( DiagonalQuadraticForm(K, [-ci, C[1], C[2], C[3]]) )
                    and IsIsotropic( DiagonalQuadraticForm(K, [ci, C[4], C[5]]) ) then
                    Append(~cc, ci);
                    break;
                end if;
            end for;
        end for;
        assert #HH eq #cc;
        
        // vector beta
        beta := [ FF2 | (b eq -1) select 1 else 0 : b in beta_ ] where 
            beta_ is &cat [ [ HilbertSymbol(cc[i], h, S[i]) : h in HH[i] ] : i in [1..#S] ];
        
        // matrix A
        // signs at the real places, where either q1 or q2 is anisotropic
        A := Matrix(FF2, [[ a eq -1 select 1 else 0 : a in RealSigns(kappa) ] : kappa in B_K ]);
        A := Matrix(Rows(Transpose(A))[inx]);
        
        // matrix B
        // Hilbert symbols
        B_ := &cat [ 
            // this is a single block B_i
            [
                [ 
                    HilbertSymbol(h, kappa, S[i]) eq -1 select 1 else 0 
                    : kappa in B_K
                ]
                : h in HH[i]
            ]
            : i in [1..#S] 
        ];
        B := Matrix(FF2, B_);
        
        // combine the matrices 
        AB := VerticalJoin(A,B);
        // combine the three vectors
        ab := Vector(FF2, alpha cat beta);

        
        // check if there is a solution
        found, Xi, _ := IsConsistent( Transpose(AB), ab);
        
        if found then
            vprint DQF_verbosity : "Xi =", Xi;
            // the sought element
            c := ProductRepresentation(B_K, Xi_) where Xi_ := Eltseq(ChangeRing(Xi,ZZ));
        else
            it, P := next_prime(it);
            while P in S do
                it, P := next_prime(it);
            end while;
            Append(~S, P);
        end if;       
    end while;
    
    // find isotropic vector of <-c> + q1
    w1 := Eltseq( isotropic_vector_NF(DiagonalQuadraticForm(K, [-c]) + q1 : alg := alg) );
    // find isotropic vector of <c> + q2
    w2 := Eltseq( isovec3_NF_builtin(DiagonalQuadraticForm(K, [c]) + q2) );

    v := [ w1[2]/w1[1], w1[3]/w1[1], w1[4]/w1[1], w2[2]/w2[1], w2[3]/w2[1] ];
    assert IsZero(Evaluate(q, v));
    return Vector(K, v);    
end function;

function isovec6_NF_NR( q : alg := "v2" )
    assert Dimension(q) eq 6;
    
    K := BaseField(q);
    OK := MaximalOrder(K);
    C := Coefficients(q);
    
    // split q at half
    q1 := DiagonalQuadraticForm(K, C[ [1,2,3] ]);
    q2 := DiagonalQuadraticForm(K, [ -C[4],-C[5],-C[6] ]);
    
    if IsIsotropic(q1) then
        w := Eltseq( isotropic_vector_NF(q) );
        v := Vector(K, w cat [0,0,0]);
        return v;
    end if;
    if IsIsotropic(q2) then
        w := Eltseq( isotropic_vector_NF(q) );
        v := Vector(K, [0,0,0] cat w);
        return v;
    end if;
    
    T := primes_that_matter(q);
    S := [p : p in T | (K!2 in p) or (not IsIsotropic(q1, p)) or (not IsIsotropic(q2, p)) ];
    
    cc := [];
    d1 := Determinant(q1);
    d2 := Determinant(q2);
    for p in S do
        V, mV := LocalMultiplicativeGroupModSquares( p );
        for v in V do
            ci := OK!mV(v);
            if not IsLocalSquare(-ci*d1,p) and not IsLocalSquare(-ci*d2,p) then
                Append(~cc, ci);
                break;
            end if;
        end for;
    end for;
    assert #cc eq #S;
    
    c := ChineseRemainderTheorem( cc, [ p^(1+Valuation(K!4,p)) : p in S] );
    // find isotropic vector of <-c> + q1
    q1_ := DiagonalQuadraticForm(K, [-c]) + q1;
    w1 := Eltseq( isotropic_vector_NF(q1_ : alg := alg) );
    // find isotropic vector of <c> + q2
    q2_ := DiagonalQuadraticForm(K, [-c]) + q2;
    w2 := Eltseq( isotropic_vector_NF(q2_ : alg := alg) );

    v := [ w1[2]/w1[1], w1[3]/w1[1], w1[4]/w1[1], w2[2]/w2[1], w2[3]/w2[1], w2[4]/w2[1] ];
    assert IsZero(Evaluate(q, v));
    return Vector(K, v);    
end function;

function isovec6_NF_R( q : alg := "v2" )
    assert Dimension(q) eq 6;
    
    K := BaseField(q);
    //OK := MaximalOrder(K);
    C := Coefficients(q);
    
    // split q at half
    q1 := DiagonalQuadraticForm(K, C[ [1,2,3] ]);
    q2 := DiagonalQuadraticForm(K, [ -C[4],-C[5],-C[6] ]);
    
    if IsIsotropic(q1) then
        w := Eltseq( isotropic_vector_NF(q) );
        v := Vector(K, w cat [0,0,0]);
        return v;
    end if;
    if IsIsotropic(q2) then
        w := Eltseq( isotropic_vector_NF(q) );
        v := Vector(K, [0,0,0] cat w);
        return v;
    end if;
    
    T := primes_that_matter(q);
    S := [p : p in T | (K!2 in p) or (not IsIsotropic(q1, p)) or (not IsIsotropic(q2, p)) ];
    
    cc := [];
    d1 := Determinant(q1);
    d2 := Determinant(q2);
    for p in S do
        V, mV := LocalMultiplicativeGroupModSquares( p );
        for v in V do
            ci := mV(v);
            if not IsLocalSquare(-ci*d1,p) and not IsLocalSquare(-ci*d2,p) then
                Append(~cc, ci);
                break;
            end if;
        end for;
    end for;
    assert #cc eq #S;
    
    a := positive_approximation(S, [1+Valuation(K!4,p) : p in S], cc );
    // real places of K
    Soo := RealPlaces(K);
    sgn_q := [ RealSigns(c) : c in C ];
    sgn_q1 := [ sgn_q[1][i] + sgn_q[2][i] + sgn_q[3][i] : i in [1..#Soo] ];
    sgn_q2 := [ sgn_q[4][i] + sgn_q[5][i] + sgn_q[6][i] : i in [1..#Soo] ];
    // find the places where at least one of q1, q2 is anisotropic
    inx := [ i : i in [1..#Soo] | (Abs(sgn_q1[i]) eq 3) or (Abs(sgn_q2[i]) eq 3) ];
    
    signs1 := signatures_builtin(q1);
    signs2 := signatures_builtin(q2);
    I := [ 
        signs1[j] eq 3 select 1 else (
        signs1[j] eq -3 select -1 else (
        signs2[j] gt 0 select -1 else 1 ))
        : j in [1..#Soo] 
    ];

    b := strong_ordering_separation_v3(K, I, S);
    c := a*b;
    
    // find isotropic vector of <-c> + q1
    q1_ := DiagonalQuadraticForm(K, [-c]) + q1;
    w1 := Eltseq( isotropic_vector_NF(q1_ : alg := alg) );
    // find isotropic vector of <c> + q2
    q2_ := DiagonalQuadraticForm(K, [-c]) + q2;
    w2 := Eltseq( isotropic_vector_NF(q2_ : alg := alg) );

    v := [ w1[2]/w1[1], w1[3]/w1[1], w1[4]/w1[1], w2[2]/w2[1], w2[3]/w2[1], w2[4]/w2[1] ];
    assert IsZero(Evaluate(q, v));
    return Vector(K, v);    
end function;




function isovec7up_NF_NR( q : alg := "v2" )
// Computes an isotropic vector of a quadratc form of dimension 7 or above over a
// non-real number field
    assert Dimension(q) gt 6;
    
    K := BaseField(q);
    C := Coefficients(q);
    q_ := DiagonalQuadraticForm(K, C[ [1..6] ]);
    w := Eltseq( isotropic_vector_NF(q_ : alg := alg) );
    v := w cat [ 0 : j in [7..Dimension(q)] ];
    
    assert IsZero(Evaluate(q, v));
    return Vector(K, v);    
end function;



function isovec7_NF_R( q : alg := "v2" )
// Computes an isotropic vector of a quadratc form of dimension 7 over a
// formally real number field
    assert Dimension(q) eq 7;
    
    K := BaseField(q);
    //OK := MaximalOrder(K);
    C := Coefficients(q);
    
    // split q at half
    q1 := DiagonalQuadraticForm(K, C[ [1,2,3] ]);
    q2 := DiagonalQuadraticForm(K, [ -C[4],-C[5],-C[6],-C[7] ]);
    
    if IsIsotropic(q1) then
        w := Eltseq( isotropic_vector_NF(q) );
        v := Vector(K, w cat [0,0,0,0]);
        return v;
    end if;
    if IsIsotropic(q2) then
        w := Eltseq( isotropic_vector_NF(q) );
        v := Vector(K, [0,0,0] cat w);
        return v;
    end if;
    
    T := primes_that_matter(q);
    S := [p : p in T | (K!2 in p) or (not IsIsotropic(q1, p)) ];
    
    cc := [];
    d1 := Determinant(q1);
    for p in S do
        V, mV := LocalMultiplicativeGroupModSquares( p );
        for v in V do
            ci := mV(v);
            if not IsLocalSquare(-ci*d1,p) then
                Append(~cc, ci);
                break;
            end if;
        end for;
    end for;
    assert #cc eq #S;
    assert2 forall { i : i in [1..#cc] | not IsLocalSquare(-cc[i]*d1,S[i]) };
    
    a := positive_approximation(S, [1+Valuation(K!4,p) : p in S], cc );
    // real places of K
    Soo := RealPlaces(K);
    sgn_q := [ RealSigns(c) : c in C ];
    sgn_q1 := [ sgn_q[1][i] + sgn_q[2][i] + sgn_q[3][i] : i in [1..#Soo] ];
    sgn_q2 := [ sgn_q[4][i] + sgn_q[5][i] + sgn_q[6][i] : i in [1..#Soo] ];
    // find the places where at least one of q1, q2 is anisotropic
    inx := [ i 
        : i in [1..#Soo] 
        | (Abs(sgn_q1[i]) eq Dimension(q1)) or (Abs(sgn_q2[i]) eq Dimension(q2)) 
    ];
    
    signs1 := signatures_builtin(q1);
    signs2 := signatures_builtin(q2);
    I := [ 
        signs1[j] eq 3 select 1 else (
        signs1[j] eq -3 select -1 else (
        signs2[j] gt 0 select -1 else 1 ))
        : j in [1..#Soo] 
    ];

    b := strong_ordering_separation_v3(K, I, S);
    c := a*b;
    
    // find isotropic vector of <-c> + q1
    q1_ := DiagonalQuadraticForm(K, [-c]) + q1;
    w1 := Eltseq( isotropic_vector_NF(q1_ : alg := alg) );
    // find isotropic vector of <c> + q2
    q2_ := DiagonalQuadraticForm(K, [-c]) + q2;
    w2 := Eltseq( isotropic_vector_NF(q2_ : alg := alg) );

    v := [ w1[2]/w1[1], w1[3]/w1[1], w1[4]/w1[1], w2[2]/w2[1], w2[3]/w2[1], w2[4]/w2[1], w2[5]/w2[1] ];
    assert IsZero(Evaluate(q, v));
    return Vector(K, v);    
end function;

function isovec8up_NF_R( q : alg := "v2" )
// Computes an isotropic vector of a quadratc form of dimension 7 over a
// formally real number field
    assert Dimension(q) eq 7;
    
    K := BaseField(q);
    //OK := MaximalOrder(K);
    C := Coefficients(q);
    
    // split q at half
    d := Dimension(q);
    k := Floor( d/2 );
    q1 := DiagonalQuadraticForm(K, C[ [1..k] ]);
    q2 := DiagonalQuadraticForm(K, [ -c : c in C[ [k+1..d] ] ]);
    
    if IsIsotropic(q1) then
        w := Eltseq( isotropic_vector_NF(q) );
        v := Vector(K, w cat [0 : j in [k+1..d]]);
        assert IsZero(Evaluate(q, v));
        return v;
    end if;
    if IsIsotropic(q2) then
        w := Eltseq( isotropic_vector_NF(q) );
        v := Vector(K, [0 : j in [1..k]] cat w);
        assert IsZero(Evaluate(q, v));
        return v;
    end if;
    
    
    // real places of K
    Soo := RealPlaces(K);
    sgn_q := [ RealSigns(c) : c in C ];
    sgn_q1 := [ sgn_q[1][i] + sgn_q[2][i] + sgn_q[3][i] : i in [1..#Soo] ];
    sgn_q2 := [ sgn_q[4][i] + sgn_q[5][i] + sgn_q[6][i] : i in [1..#Soo] ];
    // find the places where at least one of q1, q2 is anisotropic
    inx := [ i 
        : i in [1..#Soo] 
        | (Abs(sgn_q1[i]) eq Dimension(q1)) or (Abs(sgn_q2[i]) eq Dimension(q2)) 
    ];
    
    signs1 := signatures_builtin(q1);
    signs2 := signatures_builtin(q2);
    I := [ 
        signs1[j] eq 3 select 1 else (
        signs1[j] eq -3 select -1 else (
        signs2[j] gt 0 select -1 else 1 ))
        : j in [1..#Soo] 
    ];

    c := OrderingSeparation(K, I);
    
    // find isotropic vector of <-c> + q1
    q1_ := DiagonalQuadraticForm(K, [-c]) + q1;
    w1 := Eltseq( isotropic_vector_NF(q1_ : alg := alg) );
    // find isotropic vector of <c> + q2
    q2_ := DiagonalQuadraticForm(K, [-c]) + q2;
    w2 := Eltseq( isotropic_vector_NF(q2_ : alg := alg) );

    v := [ w/w1[1] : w in w1[ [2..#w1] ] ] cat [ w/w2[1] : w in w2[ [2..#w2] ] ];
    assert IsZero(Evaluate(q, v));
    return Vector(K, v);    
end function;

function isotropic_vector_NF( q : alg := "v2" )
    assert IsIsotropic(q);
    K := BaseField(q);
    
    d := Dimension(q);
    C := Coefficients(q);
    for j in [1..d] do
        q_ := DiagonalQuadraticForm(K, C[ [1..j-1] cat [j+1..d] ]);
        if IsIsotropic(q_) then
            w := Eltseq(isotropic_vector_NF(q_ : alg := alg));
            v := Vector(K, w[ [1..j-1] ] cat [0] cat w[ [j..d-1] ]);
            assert IsZero(Evaluate(q, v));
            return Vector(K, v);    
        end if;
    end for;
    
    case d :
    when 2 : return isovec2_NF(q);
    when 3 : return isovec3_NF_builtin(q);
    when 4 : return alg eq "v1" select isovec4_NF_v1(q) else isovec4_NF_v2(q);
    when 5 : return isovec5_NF(q : alg := alg);
    when 6 : 
        if IsReal(K) then
            return  isovec6_NF_R(q : alg := alg);
        else 
            return isovec6_NF_NR(q : alg := alg);
        end if;
    when 7 : 
        if IsReal(K) then
            return isovec7_NF_R(q : alg := alg); 
        else 
            return isovec7up_NF_NR(q : alg := alg);
        end if;
    else : 
        if IsReal(K) then
            return isovec8up_NF_R(q : alg := alg); 
        else 
            return isovec7up_NF_NR(q : alg := alg);
        end if;
    end case;
end function;

/*
P<x> := PolynomialRing(Rationals());
K<t> := NumberField(x^2 + 65);
q := DiagonalQuadraticForm(K, [1,-t,7-t,-t-11]);
q := DiagonalQuadraticForm(K, [-6*t + 1, 10*t - 8, 3*t - 5, -10*t + 4, 8, 6*t - 10]);
print "Isotropic:", IsIsotropic(q);

function randform( d : s := 10 )
    return DiagonalQuadraticForm(K, [ K![Random(-s, s), Random(-s, s)] : j in [1..d] ]);
end function;

for d in [2..8] do
    print "Dimension:", d;
    for j in [1..50] do
        q := randform(d);
        if not IsNondegenerate(q) then
            continue;
        end if;
        if not IsIsotropic(q) then
            continue;
        end if;
        time v := isotropic_vector_NF(q);
        
        assert Evaluate(q, v) eq 0;
    end for;
end for;


*/
