freeze;

// TODO: LegendreSymbol and IsLocalSquare in characteristic 2

import "../helpers/varia.m" : ZZ, QQ;

function primes_that_matter( q )
    if not assigned q`primes then
        q`primes := &join [ Seqset(Support(Divisor(c))) : c in q`coeffs ];
        q`primes := [ P : P in q`primes ];
    end if;
    return q`primes;
end function;

//////////////////////////////////////////////////
// HasseMinkowskiInvariant
// WittInvariant

intrinsic HasseMinkowskiInvariant( q::DiagQuadFrm, P::PlcFunElt ) -> RngIntElt
{
The Hasse invariant of a diagonal quadratic form q = < a1, ..., ad> \
at a prime P is a product of Hilbert symbols (c.f. [Lam (2005), Def. V.3.17]) \
prod_(i<j) (ai, aj)_p.
This function computes the Hasse symbol of q at p.
}
    require q`base_ring eq FunctionField(P): "Base field of P does not match base field of q!";

    d := Dimension(q);
    L := [ HilbertSymbol(q`coeffs[i],q`coeffs[j], P) : j in [(i+1)..d] , i in [1..d-1] ];
    return &* L;
end intrinsic;

intrinsic HasseMinkowskiInvariant( q::DiagQuadFrm, P::RngFunOrdIdl ) -> RngIntElt
{ " } //"
    require IsPrime(P): "P must be a prime ideal!";
    return HasseMinkowskiInvariant(q, Place(P));
end intrinsic;

intrinsic WittInvariant( q::DiagQuadFrm, P::PlcFunElt ) -> RngIntElt
{ " } //"
    return HasseMinkowskiInvariant(q, P);
end intrinsic;

intrinsic WittInvariant( q::DiagQuadFrm, P::RngFunOrdIdl ) -> RngIntElt
{ " } //"
    require IsPrime(P): "P must be a prime ideal!";
    return HasseMinkowskiInvariant(q, Place(P));
end intrinsic;
    
//////////////////////////////////////////////////
// IsIsotropic

intrinsic IsIsotropic( q::DiagQuadFrm, P::PlcFunElt ) -> BoolElt
{
Checks if a quadratic form q over a global function field is 
locally isotropic at P.
}
    require q`base_ring eq FunctionField(P): "Base field of P does not match base field of q!";

    // unary forms cannot be isotropic
    if Dimension(q) le 1 then
        return false;
    end if;
    // quintic and higher dimensional forms are always isotropic
    if Dimension(q) ge 5 then
        return true;
    end if;

    // the base field and its maximal order
    K := q`base_ring;
  
    // the residue field
    Kp, kappa := ResidueClassField(P);
    // uniformizer
    p := LocalUniformizer(P);

    // split the form into q = q0 + p*q1
    q0 := [ a*p^(-Valuation(a,P)) : a in q`coeffs | IsEven(Valuation(a,P)) ];
    q1 := [ a*p^(-Valuation(a,P)) : a in q`coeffs | IsOdd(Valuation(a,P)) ];
    //q0 := [ a*Denominator(a)^2 : a in q0 ];
    //q1 := [ a*Denominator(a)^2 : a in q1 ];
    
    // take the reduction mod gp
    q0bar := [ Evaluate(a, P) : a in q0 ];
    q1bar := [ Evaluate(a, P) : a in q1 ];
    // verify if q0, q1 are isotropic
    A := IsIsotropic( DiagonalQuadraticForm(Kp, q0bar) );
    B := IsIsotropic( DiagonalQuadraticForm(Kp, q1bar) );
    return A or B;
end intrinsic;

intrinsic IsIsotropic( q::DiagQuadFrm, P::RngFunOrdIdl ) -> BoolElt
{ " } //"
    require IsPrime(P): "P must be a prime ideal!";
    return IsIsotropic(q, Place(P));
end intrinsic;

function is_isotropic_RFF( q )
    // Given a quadratic form q with coefficients in a rational function field K
    // over a finite field, this function tests if q is isotropic.

    // the base field
    K := q`base_ring;

    // the dimension
    d := Dimension(q);
    if d le 1 then
        return false;
    elif d eq 2 then
        t, _ := IsSquare(Discriminant(q));
        return t;
    end if;

    // find all the primes of K that matter for q
//     if not assigned q`primes then
//         q`primes := &join [ Seqset(Support(Divisor(c))) : c in q`coeffs ];
//         q`primes := [ P : P in q`primes ];
//     end if;
    PP := primes_that_matter(q);
    vprint DQF_verbosity, 2 : "Primes to be tested:", PP;
    // check that q is locally isotropic over all of them
    return forall{ P : P in PP | IsIsotropic(q, P) };
end function;


//////////////////////////////////////////////////
// IsHyperbolic

intrinsic IsHyperbolic( q::DiagQuadFrm, P::PlcFunElt ) -> BoolElt
{
Given a quadratic form q over a global function field K and a prime P of K, 
this function returns true iff q is hyperbolic in the completion
of K at P.
}
    d := Dimension(q);
    // hyperbolic forms are always even-dimensional
    if IsOdd(d) then
        return false;
    end if;
    // and their discriminants are squares
    if not IsLocalSquare(Discriminant(q), P) then
        return false;
    end if;
    m := ZZ ! (d/2);
    K := q`base_ring;
    t := HilbertSymbol(K!-1,K!-1,P)^( ZZ ! (m^2-m)/2 );
    return HasseMinkowskiInvariant(q, P) eq t;
end intrinsic;

intrinsic IsHyperbolic( q::DiagQuadFrm, P::RngFunOrdIdl ) -> BoolElt
{ " } //"
    require IsPrime(P): "P must be a prime ideal!";
    require q`base_ring eq NumberField(Order(P)): "Base field of P does not match base field of q!";
    return IsHyperbolic(q, Place(P));
end intrinsic;

function is_hyperbolic_RFF( q )
    // Given a quadratic form q over a number field K, 
    // this function returns true if q is hyperbolic and false otherwise.
    
    // the base field and its maximal order
    K := q`base_ring;

    d := Dimension(q);
    // hyperbolic forms are always even-dimensional
    if IsOdd(d) then
        return false;
    end if;
    // and their discriminants are squares
    if not IsSquare(Discriminant(q)) then
        return false;
    end if;

    // find all the primes of K that matter for q
//     if not assigned q`primes then
//         q`primes := &join [ Seqset(Support(Divisor(c))) : c in q`coeffs ];
//         q`primes := [ P : P in q`primes ];
//     end if;
    PP := primes_that_matter(q);
    vprint DQF_verbosity, 2 : "Primes to be tested:", PP;
    // check that q is locally isotropic over all of them
    return forall{ P : P in PP | IsHyperbolic(q, P) };
end function;

//////////////////////////////////////////////////
// AnisotropicDimension

intrinsic AnisotropicDimension( q::DiagQuadFrm, P::PlcFunElt ) -> RngIntElt
{
Every non-degenerate quadratic form q decomposes (uniquely up to an isometry) 
as q ~ q_1 + ... + c*(1, -1), where q_1 is anisotropic. The form q_1 is called an anisotropic part of q.

Given a quadratic form q and a prime P, this function returns a dimension 
of an anisotropic part of localilization of q at P.
}
    require q`base_ring eq FunctionField(P): "Base field of P does not match base field of q!";
    
    d := Dimension(q);
    if IsEven(d) then
        // dimension of hyperbolic form is 0
        if IsHyperbolic(q, P) then
            return 0;
        elif IsLocalSquare(Discriminant(q), P) then
            return 4;
        else 
            return 2;
        end if;
    end if;
    // when q is odd-dimensional, extend it to the next even dimension
    c := QQ ! (-1)^(ZZ ! (d*(d+1)/2));
    q2 := DiagonalQuadraticForm( q`base_ring, q`coeffs cat [ c * &* q`coeffs ] );
    if IsHyperbolic( q2, P ) then
        return 1;
    else
        return 3;
    end if;
end intrinsic;

intrinsic AnisotropicDimension( q::DiagQuadFrm, P::RngFunOrdIdl ) -> RngIntElt
{ " } // "
    require IsPrime(P): "P must be a prime ideal!";
    return AnisotropicDimension(q, Place(P));
end intrinsic;

function anisotropic_dimension_RFF( q )
// Every non-degenerate quadratic form q decomposes (uniquely up to an isometry) 
// as q ~ q_1 + ... + c*(1, -1), where q_1 is anisotropic. 
// The form q_1 is called an anisotropic part of q.
//
// Given a quadratic form q over a global function field K, 
// this function returns a dimension of an anisotropic part of q.

    // the base field
    K := q`base_ring;

    // find all the primes of K that matter for q
//     if not assigned q`primes then
//         q`primes := &join [ Seqset(Support(Divisor(c))) : c in q`coeffs ];
//         q`primes := [ P : P in q`primes ];
//     end if;
    PP := primes_that_matter(q);
    vprint DQF_verbosity, 2 : "Primes to be tested:", PP;
    // compute maximal anisotropic dimension over all primes
    M := Max([ AnisotropicDimension(q, P) : P in PP ]);
    return M;
end function;


//////////////////////////////////////////////////
// IsTorsion
// just for completeness
function is_torsion_RFF( _q )
// Check if the element _q of a Witt ring is torsion.
    return true;
end function;





